//*BHEADER* :ts=8  -*- C++ -*-
/*****************************************************************************
 *
 *   |_|_|_  |_|_    |_    |_|_|_  |_		     C O M M U N I C A T I O N
 * |_        |_  |_  |_  |_        |_		               N E T W O R K S
 * |_        |_  |_  |_  |_        |_		                     C L A S S
 *   |_|_|_  |_    |_|_    |_|_|_  |_|_|_|_	                 L I B R A R Y
 *
 * $Id: HashStatic.c,v 0.29 1995/07/05 12:47:20 cncl-adm Exp $
 *
 * CNClass: CNHashStatic --- hash tables with static number of entries
 *
 *****************************************************************************
 * Copyright (C) 1992-1995   Communication Networks
 *                           Aachen University of Technology
 *                           D-52056 Aachen
 *                           Germany
 *                           Email: cncl-adm@dfv.rwth-aachen.de
 *****************************************************************************
 * This file is part of the CN class library. All files marked with
 * this header are free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.  This library is
 * distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
 * License for more details.  You should have received a copy of the GNU
 * Library General Public License along with this library; if not, write
 * to the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 **EHEADER********************************************************************/

#include "HashStatic.h"

CNHashStatic::CNHashStatic(unsigned long cap)
{
    init_table(cap);
}

CNHashStatic::CNHashStatic(CNParam *)
{
    init_table(DEFAULT_HASH_TABLE_CAPACITY);
}

CNHashStatic::~CNHashStatic()
{
    delete (table);
}

void CNHashStatic::store_key(CNKey *k)
{
    if (k == NIL) 
	fatal("You must provide a valid key, when trying to store one!\n");

    hash(k, k->hash(capacity));
}

void CNHashStatic::hash(CNKey *k, unsigned long hash_value)
{
    unsigned long pos = hash_value;
    
    if (pos > capacity) 
	fatal("hash() has been passed a hash value,\n",
	      "which exceeds the hash table's capacity!\n");
    
    if (entries == capacity)
	fatal("Unable to store object in already full hash table!\n");

    // collision?
    if (table[pos].he_CNKey != NIL)
	pos = collision(pos);

    table[pos].he_CNKey = k;
    table[pos].he_HashValue = hash_value;
    entries++;
}

CNKey *CNHashStatic::get_key(const CNKey *k) const
{
    unsigned long pos;
    
    if (k == NIL)
	fatal("You must provide a valid key, in order to get a key\n");

    pos = find_key(k, k->hash(capacity));

    // find_key failed
    if (pos == capacity)
	return NIL;
    
    return table[pos].he_CNKey;
}

unsigned long CNHashStatic::find_key(const CNKey *k, unsigned long hash_value) const
{
    unsigned long i, pos;
    
    if (k == NIL)
	fatal("You must provide a valid key, when trying to find a key\n");
    
    // quadratic search
    for (i = 0; i <= capacity - 1; i++) {
	pos = (hash_value + i * i) % capacity;
	
	// hash table does not contain the key searched for
	if (table[pos].he_CNKey == NIL) 
	    return capacity;
	// key found
	else if (table[pos].he_CNKey->compare(k)) 
	    return pos;
    }
    
    // unable to find key via quadratic searching
    // use linear searching
    for (i = 1; i <= capacity - 1; i++) {
	pos = (hash_value + i) % capacity;
	// hash table does not contain the key searched for
	if (table[pos].he_CNKey == NIL) 
	    return capacity;
	// key found
	else if (table[pos].he_CNKey->compare(k)) 
	    return pos;
    }

    // unable to find key
    return capacity;
}

CNObject *CNHashStatic::get_object(const CNKey *k) const
{
    CNKey *key;
    
    key = get_key(k);
    if (key != NIL)
	return key->get_object();
    
    return NIL;
}

bool CNHashStatic::reset()
{
    unsigned long i;

    if (!entries) {
	warning("Unable to reset() already empty hash table!\n");
	return (FALSE);
    }
    
    entries = 0;
    for (i = 0; i < capacity; i++)
	table[i].he_CNKey = NIL;

    return (TRUE);
}

bool CNHashStatic::reset_absolutely()
{
    unsigned long i;

    if (entries) 
	for (i = 0; i < capacity; i++)
	    delete table[i].he_CNKey;

    return reset();
}

bool CNHashStatic::delete_key(const CNKey *k)
{
    return remove_key(k, FALSE);
}

bool CNHashStatic::delete_key_absolutely(const CNKey *k)
{
    return remove_key(k, TRUE);
}

bool CNHashStatic::remove_key(const CNKey *k, bool remove_absolutely)
{
    unsigned long i, pos;
    struct HashEntry *old_table, *help_table;

    if (k == NIL)
	fatal("You must provide a valid key, in order to delete a key\n");
    
    if ((pos = find_key(k, k->hash(capacity))) == capacity) {
	warning("Unable to delete object: CNObject not found!");
	return FALSE;
    }

    // delete entry
    if (remove_absolutely)
	delete table[pos].he_CNKey;

    table[pos].he_CNKey = NIL;
    entries = 0;

    old_table = new (struct HashEntry[capacity]);
    if (old_table == NIL) 
	fatal("Out of memory!\n");

    // init hash table	 
    for (i = 0; i < capacity; i++)
	old_table[i].he_CNKey = NIL;

    // exchange
    help_table = old_table;
    old_table = table;
    table = help_table;
    
    for (i = 0; i < capacity; i++) 
	if (old_table[i].he_CNKey != NIL) 
	    hash(old_table[i].he_CNKey, old_table[i].he_HashValue);

    delete old_table;
	    
    return TRUE;
}


/***** Private member functions **********************************************/
unsigned long CNHashStatic::collision(unsigned long hash_value)
{
    unsigned long i, pos = hash_value;

    // calc new hash value
    // first try quadratic probing
    for (i = 1; i <= capacity - 1; i++) {
	pos = (hash_value + i * i) % capacity;
	// found a free position
	if (table[pos].he_CNKey == NIL) 
	    return (pos);
    }
    
    // unable to find a free postion via quadratic probing
    // use linear probing, which inevitably leads to clustering
    for (i = 1; i <= capacity - 1; i++) {
	pos = (hash_value + i) % capacity;
	// found a free position
	if (table[pos].he_CNKey == NIL) 
	    break;
    }
    
    return pos;
}

void CNHashStatic::init_table(unsigned long cap)
{
    const unsigned long NumPrimes = 19;
    const unsigned long Primes[NumPrimes] = 
        { 3, 7, 17, 37, 79, 163, 331, 673, 1361,
	  2729, 5471, 10949, 21911, 43853, 87719,
	  175447, 350899, 701819, 999983 
	};
    unsigned long i;
    
    if (cap != DEFAULT_HASH_TABLE_CAPACITY) {
	// find next prime number
	for (i = 0; i < NumPrimes; i++) {
	    if (cap <= Primes[i]) {
		capacity = Primes[i];
		break;
	    }
	}
	// capacity exceeds largest known prime number
	if (i == NumPrimes)
	    capacity = cap;
    } else
	capacity = cap;
    
    table = new (struct HashEntry[capacity]);
    if (table == NIL) 
	fatal("Out of memory!");

    // init hash table	 
    for (i = 0; i < capacity; i++)
	table[i].he_CNKey = NIL;

    entries = 0;
}

/***** Default I/O member function for CNCL classes **************************/

// CNNormal output
void CNHashStatic::print(ostream &strm) const
{
    strm << "capacity " << capacity
	 << " entries " << entries
	 << endl;
}

// Debug output
void CNHashStatic::dump(ostream &strm) const
{
    strm << "CNHashStatic { $Revision: 0.29 $"
	 << " capacity " << capacity
	 << " entries " << entries	 
	 << " }" << endl;
}

/***** CNCL stuff for type information and exemplar objects ******************/

// Describing object for class CNHashStatic
static CNClass CNHashStatic_desc("CNHashStatic", "$Revision: 0.29 $",
			  CNHashStatic::new_object);

// "Type" for type checking functions
CNClassDesc CN_HASHSTATIC = &CNHashStatic_desc;






